SuperStrict
Rem
	bbdoc: Crypting Streams.
	about: A Stream Wrapper providing simple crypt mechanisms.
End Rem
Module ZEVS.CryptStream
Import BRL.Stream
Import BRL.Random


Private
Type TCryptStream Extends TStreamWrapper
  Field rnd_state%, crypted% = True, tmp%
  Function Create:TCryptStream(stream:TStream, key$)
    Local result:TCryptStream = New TCryptStream
    result.SetStream stream
    result.rnd_state = Self._Hash(key)
    Return result
  End Function
  Method SetCrypt(bool%)
    Self.crypted = crypted
  End Method
  Method SetKey(key$)
    Self.rnd_state = Self._Hash(key)
  End Method
  Method Read%(buf:Byte Ptr, count%)
    Local c% = Super.Read(buf, count)
    Self._CryptBytes buf, c
    Return c
  End Method
  Method Write%(buf:Byte Ptr, count%)
    Local mem:Byte Ptr = MemAlloc(count)
    Assert mem
    MemCopy mem, buf, count
    Self._CryptBytes mem, count
    Local c% = Super.Write(mem, count)
    MemFree mem
    Return c
  End Method
  Method ReadByte%()
    If Self.crypted Then Return Self._CryptByte(Super.ReadByte())
    Return Super.ReadByte()
  End Method
  Method WriteByte%(n%)
    If Self.crypted Then Return Super.WriteByte(Self._CryptByte(n))
    Return Super.WriteByte(n)
  End Method
  Method ReadShort%()
    Local n% = Super.ReadShort()
    Self._CryptBytes Varptr n, 2
    Return n
  End Method
  Method WriteShort%(n%)
    Self._CryptBytes Varptr n, 2
    Return Super.WriteShort(n)
  End Method
  Method ReadInt%()
    Local n% = Super.ReadInt()
    Self._CryptBytes Varptr n, 4
    Return n
  End Method
  Method WriteInt%(n%)
    Self._CryptBytes Varptr n, 4
    Return Super.WriteInt(n)
  End Method
  Method ReadFloat#()
    Local n# = Super.ReadFloat()
    Self._CryptBytes Varptr n, 4
    Return n
  End Method
  Method WriteFloat%(n#)
    Self._CryptBytes Varptr n, 4
    Return Super.WriteFloat(n)
  End Method
  Method ReadDouble!()
    Local n! = Super.ReadDouble()
    Self._CryptBytes Varptr n, 8
    Return n
  End Method
  Method WriteDouble%(n!)
    Self._CryptBytes Varptr n, 8
    Return Super.WriteDouble(n)
  End Method
  Method ReadLong:Long()
    Local l:Long
    Self.Read Varptr l, 8
    Return l
  End Method
  Method WriteLong%(n:Long)
    Return Self.Write(Varptr n, 8)
  End Method
  Method ReadLine$()
    If Not Self.crypted Then Return Super.ReadLine()
    Self._StartCrypt
    Local buf:Byte[1024],sz%
    Repeat
      Local ch:Byte
      Try
        ch = Super.ReadByte() ~ Rand(0, 255)
      Catch e:TStreamException
        Exit
      EndTry
      If ch=0 Or ch=10 Exit
      If ch=13 Continue
      If sz=buf.length buf=buf[..sz*2]
      buf[sz]=ch
      sz:+1
    Forever
    Self._EndCrypt
    Return String.FromBytes( buf,sz )
  End Method
  Method WriteLine%(str$)
    If Not Self.crypted Then Return Super.WriteLine(str)
    Self._StartCrypt
    str :+ "~r~n"
    Local mem:Byte Ptr = MemAlloc(str.length)
    For Local i%=0 Until str.length
      mem[i] = str[i] ~ Rand(0, 255)
    Next
    Local result% = Super.Write(mem, str.length)
    MemFree mem
    Self._EndCrypt
    Return result
  End Method
  Method ReadString$(n%)
    If Not Self.crypted Then Return Super.ReadString(n)
    Local mem:Byte Ptr = MemAlloc(n)
    Self.Read mem, n
    Local result$ = String.FromBytes(mem, n)
    MemFree mem
    Return result
  End Method
  Method WriteString%(str$)
    If Self.crypted Then
      Self._StartCrypt
      Local mem:Byte Ptr = MemAlloc(str.length)
      For Local i%=0 Until str.length
        mem[i] = str[i] ~ Rand(0, 255)
      Next
      str = String.FromBytes(mem, str.length)
      Self._EndCrypt
      MemFree mem
    EndIf
    Return Super.WriteString(str)
  End Method
  Method ReadObject:Object()
    Assert Not Self.crypted, "Cannot crypt objects"
    Return Super.ReadObject()
  End Method
  Method WriteObject%(obj:Object)
    Assert Not Self.crypted, "Cannot crypt objects"
    Return Super.WriteObject(obj)
  End Method




  Function _Hash%(str$)
    Local hash% = $12345678
    Local b% = 0
    For Local i%=0 Until str.length
      hash :~ str[i] Shl b
      b :+ 8
      If b = 32 Then b = 0
    Next
    Return hash
  End Function
  Method _StartCrypt()
    Self.tmp = RndSeed()
    SeedRnd Self.rnd_state
  End Method
  Method _EndCrypt()
    Self.rnd_state = RndSeed()
    SeedRnd Self.tmp
  End Method
  Method _CryptByte:Byte(b:Byte)
    Return b ~ Rand(0, 255)
  End Method
  Method _CryptBytes(buf:Byte Ptr, count%)
    If Self.crypted Then
      Self._StartCrypt
      For Local i%=0 Until count
        buf[i] :~ Rand(0, 255)
      Next
      Self._EndCrypt
    EndIf
  End Method
End Type
Rem
	bbdoc: The Factory for crypted streams.
	about: A crypted stream has the protocol <code>crypt:<i>key</i></code>. This Type is Private.
End Rem
Type TCryptStreamFactory Extends TStreamFactory
  Const PROTO$ = "crypt:"
  Method CreateStream:TStream(url:Object, proto$, path$, readable%, writeable%)
    If Not proto.StartsWith(Self.PROTO) Then Return Null
    Local key$ = proto[Self.PROTO.length..]
    Local stream:TStream = OpenStream(path, readable, writeable)
    If stream Then Return TCryptStream.Create(stream, key)
  End Method
End Type
New TCryptStreamFactory
Public
Rem
	bbdoc: Enables or disables crypting.
	about: Plain saved data is readable for everyone and must be read as plain data (StreamSetCrypt ..., False).
End Rem
Function StreamSetCrypt(stream:TStream, bool% = True)
  If Not TStreamWrapper(stream) Then RuntimeError "Stream must be crypt stream"
  If TCryptStream(stream) Then
    TCryptStream(stream).SetCrypt bool
    Return
  EndIf
  StreamSetCrypt TStreamWrapper(stream)._stream, bool
End Function
Rem
	bbdoc: Sets the crypt's key.
End Rem
Function StreamSetKey(stream:TStream, key$)
  If Not TStreamWrapper(stream) Then RuntimeError "Stream must be crypt stream"
  If TCryptStream(stream) Then
    TCryptStream(stream).SetKey key
    Return
  EndIf
  StreamSetKey TStreamWrapper(stream)._stream, key
End Function